print "Vigenere functions in Sage. Version 1.1 date: Feb 18, 2017\n"
# v1.1 a small change in the documenation
print "The main functions of interest in this file are : "
print "encrypt_vig(plaintext, key) - encrypt plaintext with the key with Vigenere"
print "decrypt_vig(cyphertext, key) - decrypt cyphertext with key with Vigenere"
print "clean_text(ptext) - break text into 5-grams and make uppercase"
print "show_all_shifts(cyphertext, position, period) - display 26 shifts of statistics"
print "break_vig(cyphertext, key_guess) - display the statistics for each letter of the key\n"

alpha = "ABCDEFGHIJKLMNOPQRSTUVWXYZ" # the English alphabet
engstats = [73,9,30,44,130,28,16,35,74,2,3,35,25,78,74,27,3,77,63,93,27,13,16,5,19,1] # and their frequency out of 1000
# this is a global variable because we use it over and over

def encrypt_vig(plaintext, key):
    """
    Encrypt with the Vigenere encypherment scheme
    
    INPUT:

    - ``plaintext`` -- a string representing the plaintext

    - ``key`` -- a string representing the key
    
    OUTPUT:

    - a string representing the cyphertext

    EXAMPLES::

        sage: encrypt_vig("NOW IS THE TIME", "DOG")
        'QCC LG ZKS ZLAK'
        sage: encrypt_vig("Now is THE time", "dog")
        'Qcc lg ZKS zlak'
    """
    nkey = [alpha.rfind(let.upper()) for let in key] # make a numeric key
    cyphertext = ""
    count = 0
    for let in plaintext:
        if let.isalpha(): # if it is a letter
            if let.isupper():
                cyphertext+=(2*alpha)[alpha.rfind(let)+nkey[mod(count,len(nkey))]]
            else:
                cyphertext+=(2*alpha.lower())[alpha.lower().rfind(let)+nkey[mod(count,len(nkey))]]
            count+=1
        else: # if it is a space or punctuation
            cyphertext+=let
    return cyphertext

def ekey_to_dkey_vig(ekey):
    """
    Return a key that can be used to decrypt Vigenere by encrypting with this key
    
    The decrypting key is just the encrypting key with the replacement 'a' <-> 'A',
    'b' <-> 'Z', 'c' <-> 'Y', 'd' <-> 'X', etc.

    INPUT:

    - ``ekey`` -- a string representing the encrypting key

    OUTPUT:

    - a string representing the decrypting key
    EXAMPLES:
        sage: ekey_to_dkey_vig("dog")
        'XMU'
        sage: ekey_to_dkey_vig("abba")
        'AZZA'
        sage: ekey_to_dkey_vig("ABCDEFG")
        'AZYXWVU'
    """
    return "".join([alpha[-alpha.rfind(let.upper())] for let in ekey])

def decrypt_vig(cyphertext, key):
    """
    Decrypt with the Vigenere encypherment scheme
    
    INPUT:

    - ``cyphertext`` -- a string representing the plaintext

    - ``key`` -- a string representing the key
    
    OUTPUT:

    - a string representing the plaintext

    EXAMPLES::

        sage: decrypt_vig("QCC LG ZKS ZLAK", "DOG")
        'NOW IS THE TIME'
        sage: decrypt_vig("Qcc lg ZKS zlak", "dog")
        'Now is THE time'
    """
    return encrypt_vig(cyphertext, ekey_to_dkey_vig(key))

def clean_text(ptext):
    """
    Change lowercase letters to upper case and delete spaces and punctuation
    put text into 5 grams

    INPUT:

    - ``ptext`` -- a string

    OUTPUT:

    - a string with only uppercase letters and spaces

    EXAMPLES::

        sage: clean_text("Now * the TIME :: for all good !")
        'NOWTH ETIME FORAL LGOOD'
    """
    out = ""
    cnt = 0
    for i in range(len(ptext)):
        if ptext[i].isalpha():
            if not mod(cnt,5) and cnt>0:
                out+=" "
            out += ptext[i].upper()
            cnt += 1
    return out

def create_counts(cyphertext, position, period):
    """
    Break the text into ``period``-grams and then count the number of letters in
    the ``position`` places in those grams.

    Spaces and non alphabetic characters are ignored

    INPUT:

    - ``cyphertext`` -- a string representing the cyphertext

    - ``position`` -- an integer between 1 and ``period``

    - ``period`` -- the length of the key

    OUTPUT:

    - a list of length 26 with the number of letters appearing in positions
    ``position`` + ``period``*k for some k

    EXAMPLES::

        sage: clean_text("Now * the TIME :: for all good !")
        'NOWTH ETIME FORAL LGOOD'
        sage: create_counts("Now * the TIME :: for all good !",1,5)
        [0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        sage: create_counts("Now * the TIME :: for all good !",5,5)
        [0, 0, 0, 1, 1, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
    """
    cnt = 0
    stats = [0]*26
    for i in range(len(cyphertext)):
        if cyphertext[i].isalpha():
            cnt+=1
            if mod(cnt,period)==position:
                stats[alpha.rfind(cyphertext[i].upper())]+=1
    return stats

def graph_stats(counts, showscore=False):
    """
    plot the statistics of the counts of the numbers of letters in blue
    next to the statistics of English in red

    INPUT:

    - ``counts`` -- a list of 26 non-negative integers

    - ``showscore`` -- a boolean

    OUTPUT:

    - a graphics object with the English statistics drawn in red and the
    statistics ``counts`` drawn in blue

    """
    G = Graphics()
    G+=line([[0,0],[260,0],[260,100],[0,100],[0,0]])
    N = sum(counts)
    score = sum(counts[i]*engstats[i] for i in range(26))
    clean = [min((1000*val)/N,180) for val in counts]
    for i in range(26):
        G+=text(alpha[i],(10*i+5,-3))
        G+=polygon2d([[10*i+5,0],[10*i+8,0],[10*i+8,engstats[i]/2],[10*i+5,engstats[i]/2],[10*i+5,0]],color='red')
        G+=polygon2d([[10*i+2,0],[10*i+5,0],[10*i+5,clean[i]/2],[10*i+2,clean[i]/2],[10*i+2,0]],color='blue')
        if showscore:
            G+=text("score: %s"%(score),(20,90),color='green')
        if clean[i]==180:
            G+=text("+",(10*i+4,95))
    return G

def show_all_shifts(cyphertext, position, period, showscore=False):
    """
    plot all of the possible shifts of the statistics

    INPUT:

    - ``cyphertext`` -- a string representing the cyphertext

    - ``position`` -- an integer between 1 and ``period``

    - ``period`` -- the length of the key

    OUTPUT:

    - None, but as a side effect it will show 26 pictures with all shifts
    of the statistics for the letters in positions ``position``+k*``period``

    """
    print "Note the English statistics are in red, cyphertext is in blue"
    counts = create_counts(cyphertext,position,period)*2
    for v in range(26):        
        (graph_stats(counts[v:26+v],showscore)+text("shift by "+alpha[v],(290,50))).show(axes=False)

def break_vig(cyphertext, key):
    """
    Given the cyphertext and a guess at the key, display the statistics
    for each of the positions of the key
    
    INPUT:

    - ``cyphertext`` -- a string representing the cyphertext

    - ``key`` -- a string representing the key
    
    OUTPUT:

    - None, but as a side effect it graphs the letter counts against the
    English letter counts for each position in the key

    EXAMPLES::

        sage: break_vig(ct, "AAAAAA")
    """
    ptextguess = decrypt_vig(cyphertext, key)
    for i in range(len(key)):
        print "statistics for positions %s of period %s :"%(i+1,len(key))
        graph_stats(create_counts(ptextguess, i+1, len(key))).show(axes=False)
